c                                                                                               
C      A Decomposed Multi-objective Genetic Algorithm for Solving the Joint Product Family
C      Platform-Selection and Variant-Design Problem with Generalized Commonality

c      Case study: universal electric motor family

C      Aida Khajavirad 
C      Department of Mechanical Engineering
C      Carnegie Mellon University

c     ---------------------------------------------------------------------------------------

C      The MOGA code finds the optimal product families with minimum performance deviaion
c      and maximum commonality value. the code is illustrated for optmizing a family of 10 
c      universal electric motors.

c      The performace objective is defined as the sum of deviations from the performance 
c      targets and the commonality value is calculated according to the commonality index  
c      introduced by Martin and Ishii (1997) which measures the number of unique parts in 
c      the family.

c      The multi-objective optmization is applied according the NSGAII scheme (Deb 2001); 
c      however,for this specific application (product family optimization), several aspects 
c      such as chromosomerepresentation, crossover and mutaion operators have been modified 
c      (Khajavirad et al., 2007)
c      For adressing the scalablity issue the model is decomposed to an upper level GA for 
c      finding the optimal platform configuration using generalized commonality (i.e. allowing
c      for component sharing within subsets of variants) and lower levels, each for optimizing 
c      an individual variant. moreover, the model has been parallelized using MPI library: 
c      total number of nodes: number of products+1 

c      The code consists of the following subroutines that are called in the main code (GA.for):
         
c	  - crossover.for: applies 2d binary crossover and simulated binary crossover operators
c                        to upper and lower levels respectively.
c       - crowdingdistance.for : calculates the crowding distance according to NSGAII model for
c                                each front member
c       - definerange.for : defines the lower and upper bounds for each design variable
c       - findfitness.for : finds the fitness of each member in all sub-GAs
c       - initialpopulation.for: intialize a random population in all sub-GAs
c       - modifypop.for: modifies the population according to consistency constraints 
c                        (khajavirad et al. 2007)
c       - mutation operator: applies 2D flipping mutation and polynomial mulataion 
c                            operators to upper and lower levels respectively.
c       - ndsorting.for: applies non-dominated sorting according to NSGAII scheme.
c       - random.for: random number generator
c       - replace the old population by the new one (khajavirad et al 2007)
c       - selection: applies crowded tournoment selection operator
c       - settarget.for: the targets and limits for performance objectives and constraints 
c         according to the motor example formulation
c       - variance.for : calculates the commonality value according to the commonality index
c       - voilation.for: calculates the constraint violation for each member
c       - openfiles.for: opens the input and output files
c       - print2file.for: writes the data to output files

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

       MODULE SHARED_DATA
	 IMPLICIT NONE
C	-------------------------- SHARED DATA-----------------------
C 
C      GA SHARED PARAMETERS  --------------------------------

	 INTEGER,PARAMETER::POPNUM=400 ! POPULATION NUMBER
       INTEGER,PARAMETER::MAXGEN=250 ! MAXIMUM GENERATION NUMBER 
	 REAL*8,PARAMETER::PC=9.0D-1   ! CROSSOVER PROBABILITY
	 REAL*8,PARAMETER::PM=1.0D-1   ! MUTATION PROBABILITY 
	 REAL*8,PARAMETER::ETAC=2.0D0  ! CROSSOVER PROBABILITY FACTOR 
	 	                           !(REF.:SIMULATED BINARY CROSSOVER OPERATOR)
	 REAL*8,PARAMETER::ETAM=5.0D0  ! MUTATION PROBABILITY FACTOR 
	                               !(REF.:POLYNOMIAL MUTATION OPERATOR)
	 REAL*8::RANGES(2,VARNUM)      ! DESIGN VARIABLES LOWER AND UPPER BOUNDS  
	 REAL*8::FITNESS1(POPNUM)      ! PERFORMANCE OBJ. FOR THE PARENTS POPULATION 
       REAL*8::NEWFITNESS1(POPNUM)   ! PERFORMANCE OBJ. FOR THE OFFSPRING POPULATION
	 REAL*8::FITNESS2(POPNUM)      ! COMMONALITY OBJ. FOR THE PARENTS POPULATION
       REAL*8::NEWFITNESS2(POPNUM)   ! COMMONALITY OBJ. FOR THE PARENTS POPULATION
	 REAL*8::DISTANCE(POPNUM)      ! CROWDING DISTANCE (REF.: NSGAII)
	 INTEGER::RANK(POPNUM)         ! NON-DOMINATED RANKING (REF.: NSGAII)
       
C      PRODUCT FAMILY SHARED PARAMETERS ----------------------------
	 
	 INTEGER,PARAMETER::PRONUM=10    ! NUMBER OF INDIVIDUAL VARIANTS 
	 INTEGER,PARAMETER::VARNUM=6     ! NUMBER OF DESIGN VARIABLES
       REAL*8::MAXCOMMONALITY=(PRONUM-1)*VARNUM ! MAXIMUM COMMONALITY VALUE
	 
     
C     PARALLEL SHARED PARAMETERS -----------------------------------
       
	 INTEGER::IRANK        ! DEFINES EACH NODE BY AN INTEGER NUMBER 
	 INTEGER::P            ! DEFINES THE TOTAL NUMBER OF NODES
       REAL*8::NEWTYP1       ! A NEW TYPE CREATED USING MPI_TYPE_VECTOR

	 
	 END MODULE

C      ===============================================================================

	 MODULE UPPERLEVEL
	 USE SHARED_DATA
	 IMPLICIT NONE

C      --------------------- THE UPPER LEVEL GA PARAMETERS -------------------

C      2D CHROMOSOME REPRESENTING GENERALIZED COMMONALITY
       
	 INTEGER::CHROM1(PRONUM,VARNUM,POPNUM)   
	 INTEGER::NEWCHROM1(PRONUM,VARNUM,POPNUM)  

C      COMMONALITY VALUE CALCULATED ACCORDING TO THE 2D CHROMOSOME
	 REAL*8::COMMVAL


	 END MODULE

C     ==================================================================================

       MODULE LOWERLEVEL
	 USE SHARED_DATA
       IMPLICIT NONE
	
C      ------------------- LOWER LEVEL GAs PARAMETERS --------------------------

C      1D CHROMOSOME CONTAINING DESIGN VARIABLES OF EACH VARIANT
       REAL*8::CHROM2(POPNUM,VARNUM)   
	 REAL*8::NEWCHROM2(POPNUM,VARNUM) 

C      PARAMETERS FOR THE UNIVERSAL MOTOR CASE STUDY
       REAL*8::TORQUE,OUT_TORQUE   !OUTPUT TORQUE
	 REAL*8::POWER,OUT_POWER     !OUTPUT POWER
	 REAL*8::EFF,MIN_EFF         !MOTOR'S EFFICIENCY
	 REAL*8::FEAS,MIN_FEAS       !FEASIBLITY VALUE
	 REAL*8::MASS,MAX_MASS       !MOTOR'S MASS
	 REAL*8::H,MAX_H            
	 REAL*8::AWA,MIN_AWA,MAX_AWA
	 REAL*8::LENGTH,MIN_LENGTH,MAX_LENGTH
	         
	 END MODULE
	
C     =====================================================================================


	 PROGRAM NSGA2
	 USE SHARED_DATA
	 USE UPPERLEVEL
	 USE LOWERLEVEL
	 IMPLICIT REAL*8(A-H,O-Z)
	 INCLUDE 'MPIF.H'


       CALL MPI_INIT(IERR)
	 CALL MPI_COMM_RANK(MPI_COMM_WORLD,IRANK,IERR)
	 CALL MPI_COMM_SIZE(MPI_COMM_WORLD,P,IERR)


       CALL MPI_TYPE_VECTOR(VARNUM,1,POPNUM,MPI_DOUBLE_PRECISION,NEWTYP1
	1                      ,IERR)
	 CALL MPI_TYPE_COMMIT(NEWTYP1,IERR)       
 


       IF (IRANK.NE.0) THEN
    	 
	    CALL DEFINERANGE    ! LOWER AND UPPER BOUNDS FOR DESIGN VARIABLES

	    CALL SETTARGET      ! TARGET AND LIMITING VALUES FOR EACH VARIANT

	 ENDIF     
	 
	 CALL OPENFILES         ! OPEN INPUT AND OUTPUT FILES

	 CALL RANDOM(IDUM,RND)  ! RANDOM NUMBER GENERATOR


C	-------------------CREATE AN INITIAL POPULATION--------------------

	 IGEN=1                      ! GENERATION NUMBER
C         
	 CALL INITIALPOPULATION      ! BUILD A RANDOM POPULATION FOR ALL SUB-GAs	 	 
	 	 
	 CALL MODIFYPOP              ! MODIFY INITIAL POPULATION ACCORDING TO 
	                                   ! CONSISTENY CONSTRAINTS
      
	 CALL FINDFITNESS(IGEN)      ! FIND EACH POPULATION MEMBER'S FITNESS
     
       CALL PRINT2FILE             ! WRITING THE DATA TO OUTPUT FILES


C	----------------MAIN LOOP OVER ALL GENERATIONS---------------------

      
	 DO IGEN=2,MAXGEN
          	    
		CALL NDSORTING         ! APPLY CONSTRAINED NON-DOMINATED SORTING 
	
		CALL CROWDINGDISTANCE  ! CALC. CROWDING DISTANCE FOR EACH FRONT'S MEMBER
	   
	    DO INUM=1,POPNUM,2

C       SELETION:  CROWDED TOURNAMENT SELECTION OPERATOR	       
	       CALL SELECTION(MATE1) 
          
      	   CALL SELECTION(MATE2) 
C       CROSSOVER: 
C           -UPPER LEVEL:  2D BINARY CROSSOVER OPERATOR
C           -LOWER LEVELS: SIMULATED BINARY CROSSOVER OPERATOR           	  
             CALL CROSSOVER(MATE1,MATE2,INUM) 
		               
	    ENDDO
C         MUTATION
C            -UPPER LEVEL: 2D FLIPPING MUTATION OPERATOR
C            -LOWER LEVEL: POLYNOMIAL MUTATION OPERATOR         
	    CALL MUTATION
          
          CALL MODIFYCROSSOVER(IGEN) ! MODIFY THE OFFSPRING POPULATION ACCORDING 
		                               !TO THE CONSISTENCY CONSTRAINTS
	    CALL FINDNEWFITNESS(IGEN)   ! FIND FITNESS OF THE OFFSPRING POPULATION
	
	    CALL REPLACE(IGEN)          ! SELECT THE NEW POPULATION ACCORDING TO 
		                            !             NSGA2 MODEL
       
	   IF (MOD(IGEN,50)==0) CALL PRINT2FILE
          
	 ENDDO
      
       	 
	 CALL PRINT2FILE       ! WRITE RESULTS TO OUPUT FILES
      
	 CALL MPI_FINALIZE(IERR)
 

	 END PROGRAM

